/***
 * ASM tests
 * Copyright (c) 2000-2011 INRIA, France Telecom
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the copyright holders nor the names of its
 *    contributors may be used to endorse or promote products derived from
 *    this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 */
package org.objectweb.asm.test.cases;

import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;

import org.objectweb.asm.Opcodes;

/**
 * Generates classes designed so that the "conform" test suite, applied to these
 * classes, covers all the ASM code base.
 * 
 * @author Eric Bruneton
 */
public class Generator implements Opcodes {

    public static void main(final String[] args) throws IOException {
        Generator generators[] = { new Annotation(), new Attribute(),
                new Debug(), new Enum(), new Frames(), new Insns(),
                new Interface(), new Invalid(), new JSR(), new Outer(),
                new Wide(), new InvokeDynamic() };
        for (int i = 0; i < generators.length; ++i) {
            generators[i].generate(args[0]);
        }
    }

    protected void generate(final String dir) throws IOException {
    }

    protected void generate(final String dir, final String path,
            final byte[] clazz) throws IOException {
        File f = new File(new File(dir), path);
        if (!f.getParentFile().exists() && !f.getParentFile().mkdirs()) {
            throw new IOException("Cannot create directory "
                    + f.getParentFile());
        }
        FileOutputStream o = new FileOutputStream(f);
        o.write(clazz);
        o.close();
    }
}
